import numpy as np
from .calculate_spkd import spkd_functions
import importlib
rs_distances_spec = importlib.util.find_spec('rs_distances')

if rs_distances_spec is not None:
    try:
        import rs_distances
    except ImportError:
        print('Error during import of rs_distances. Fallback to spkd_v function.')
        rs_distances = None
else:
    rs_distances = None

def spkd(cspks: np.ndarray | list, qvals: list | np.ndarray, use_rs: bool = True):
    """
    Compute pairwise spike train distances with variable time precision for multiple cost values.

    Args:
        cspks (nested iterable[list | np.ndarray]): Each inner list contains spike times for a single spike train.
        qvals (list of float | int): List of time precision values to use in the computation.
        use_rs (bool, optional): Whether to use the rs_distances implementation. Defaults to True.

    Returns:
        ndarray: A 3D array containing pairwise spike train distances for each time precision value.
    """
    if rs_distances is not None and use_rs:
        d = rs_distances.calculate_spkd(cspks, qvals)
        return np.maximum(d, np.transpose(d, [1, 0, 2]))
    else:
        return spkd_functions.calculate_spkd(cspks, qvals, None)


def spkd_slide(cspks: np.ndarray | list, qvals: list | np.ndarray, res: float | int = 1e-3):
    """
    Compute pairwise spike train distances with variable time precision for multiple cost values,
    incorporating sliding of one spike train along the time axis.

    Args:
        cspks (nested iterable[list | np.ndarray]): Each inner list contains spike times for a single spike train.
        qvals (list of float | int): List of time precision values to use in the computation.
        res (float, optional): The resolution of the sliding operation. Default is 1e-3.

    Returns:
        ndarray: A ND array containing pairwise spike train distances where N=len(costs), for each time precision value.
    """
    return spkd_functions.calculate_spkd(cspks, qvals, res)



import metricspace as ms
import numpy as np

# Generate random spike trains
spike_train_A = np.sort(np.random.uniform(low=0.0, high=2, size=100))
spike_train_B = np.sort(np.random.uniform(low=0.0, high=2, size=100))

# Input spike trains into a list or array (as many or few as you want)
spike_trains = [spike_train_A, spike_train_B] 

# Make array of cost values to be used in the spike-distance calculation (here we get 0 to 512)
costs = np.concatenate(([0], 2 ** np.arange(-4, 9.5, 0.5)))

spike_distance = ms.spkd(spike_trains, costs)  # Standard approach
spike_distance_slide = ms.spkd_slide(spike_trains, costs, 10e-3)  # Sliding window approach with search window of 1ms

# Cluster spike trains using spike distance and the number of samples in each class
spike_train_class_labels = np.concatenate((np.zeros(100), np.ones(100))) # 100 samples in each class, randomly generated
_, nsam = np.unique(spike_train_class_labels, return_counts=True)
clustered = ms.distclust(spike_distance, nsam)

# Calculate entropy from the confusion matrix output of distclust
mi = ms.tblxinfo(clustered)
mj = ms.tblxjabi(clustered)
mt = ms.tblxtpbi(clustered)
mij = mi + mj
mit = mi + mt


from .histbi import histbi
from .tblxbi import tblxbi
from .histtpbi import histtpbi
from .tblxtpbi import tblxtpbi
from .histinfo import histinfo
from .histjabi import histjabi
from .tblxinfo import tblxinfo

__all__ = ['histbi', 'tblxbi', 'histtpbi', 'tblxtpbi', 'histinfo', 'histjabi', 'tblxinfo']



import numpy as np


def histjabi(cvec):
    """
        Calculates the jackknife bias that should be added to the naive plugin estimate, in bits.

        Args:
            cvec (numpy.ndarray): Vector of counts.

        Returns:
            float: Jackknife bias in bits.
        """
    cvec = np.reshape(cvec, (np.prod(cvec.shape), 1))
    if np.max(cvec) <= 1:
        h = 0
        return h
    nsamps = np.sum(cvec)
    cv2 = cvec[cvec >= 2]
    jdev = np.log((nsamps - 1) / nsamps) + (1 / (nsamps * (nsamps - 1))) * np.sum(
        cv2 * (cv2 - 1) * np.log(cv2 / (cv2 - 1)))
    h = -(nsamps - 1) * jdev / np.log(2)  # first-order correction
    return h



"""
This package contains the validation functions for the metricspace package.
Helps ensure routines are expected and raises errors or warnings if undesired behavior is detected or expected. 

For example, if providing a search resolution window and not using optimizations, a warning will be raised indicating that the computation 
could take an astronomically long time. 

"""


import numpy as np


def histtpbi(cvec, useall=0):
    """
        Calculates the Treves-Panzeri bias that should be added to the naive plugin estimate, in bits.

        Args:
            cvec (numpy.ndarray): Vector of counts.
            useall (int, optional): Use all bins or specific bins based on the parameter. Default is 0.

        Returns:
            float: Treves-Panzeri bias in bits.
        """
    counts = np.reshape(cvec, (1, np.prod(cvec.shape)))
    if np.sum(counts) == 0:
        h = 0
        return h
    if useall:
        bins = len(counts)
    else:
        bins = np.sum(counts > 0)
    h = (bins - 1) / (2 * np.sum(counts) * np.log(2))
    return h



import numpy as np
from . import histinfo as hinfo


def tblxinfo(tabl):
    """
    Calculates the transinformation, in bits.

    Args:
        tabl (numpy.ndarray): 2-dimensional array of probabilities.

    Returns:
        float: Transinformation in bits.
    """
    h = hinfo.histinfo(np.sum(tabl, axis=1)) + hinfo.histinfo(np.sum(tabl, axis=0)) - hinfo.histinfo(tabl)
    return h



from . import histjabi as hjabi
from . import histtpbi as htpbi


def histbi(cvec, estimate_type, param=None):
    """
      Calculates the bias estimate for the naive plugin histogram entropy estimate, in bits.

      Args:
          cvec (numpy.ndarray): Vector of counts.
          estimate_type (str): Type of bias estimation. 'ja' for jackknife estimate, 'tp' for Treves-Panzeri estimate.
            - 'ja' uses the jackknife estimate of the bias, which is the more conservative option.
            
          param (int, optional): Additional parameter for specific bias estimation. Default is None.

      Returns:
          float: Bias estimate in bits.
      """
    h = None
    if estimate_type == 'ja':
        h = hjabi.histjabi(cvec)
    elif type == 'tp':
        useall = 0 if param is None else param
        h = htpbi.histtpbi(cvec, useall)
    return h



import numpy as np
from . import histbi as hbi
from . import tblxtpbi as ttpbi


def tblxbi(ctabl, estimate_type, param=None):
    """
        Calculates the Treves-Panzeri or jackknife bias correction to be added to the naive transinformation estimate, in bits.

        Args:
            ctabl (numpy.ndarray): 2-dimensional array of counts.
            estimate_type (str): Type of bias estimation. 'ja' for jackknife estimate, 'tr' for Treves-Panzeri estimate.
            param (int, optional): Additional parameter for specific bias estimation. Default is None.

        Returns:
            float: Bias correction estimate in bits.
        """
    h = None
    ty = estimate_type[0:2].lower()
    if ty == 'ja':
        h = hbi.histbi(np.sum(ctabl, axis=0), 'ja') + hbi.histbi(np.sum(ctabl, axis=1), 'ja') - hbi.histbi(ctabl, 'ja')
    elif ty == 'tr':
        useall = 0 if param is None else param
        h = ttpbi.tblxtpbi(ctabl, useall)
    return h



"""
Python translation of original distclust_fast.m

Translation by Flynn O'Connell, 2023.
(c) 2000 by Daniel Reich. All rights reserved.
"""
import numpy as np


def distclust(dists, nsam, expo=-2, ifresamp=0, iftrump=1):
    """
        Distance clustering. Classificatiion algorithm that returns a KxKxQ confusion matrix where K
        is the number of classes and q is the temporal coding cost metric.

        The confusion matrix is calculated by averaging the pairwise distances between trials of the same class and
        comparing them to the pairwise distances between trials of different classes. The resulting matrix is
        y_pred vs y_true. Each row should sum to the number of trials/samples in each class.

        Algorithm doesn't assume all classes have the same number of trials and
        self-distances are not included in the calculation.

        Args:
            dists (numpy.ndarray): A 2D symmetric matrix of pairwise distances with 0's on the diagonal.
                Its size should be sum(nsam) x sum(nsam).
            nsam (list or numpy.ndarray): A list or 1D array where each entry is the number of trials in each condition.
            expo (float or str): The exponent value for calculating the weighted mean distance.
                             If expo is 0, it is set to 'median'. If expo is a negative number, iftrump is set to 1.
                             If expo is 'median', the median distance is used instead of the weighted mean.
                             Default is -2 (gravitationally-weighted mean).
            ifresamp (int): The resampling method to use. Options:
                            0 - No resampling.
                            1 - Relabel resampling.
                            2 - Bootstrap resampling.
            iftrump (int): Determines whether 0-distances should trump other values. Options:
                           0 - 0-distances do not trump other values.
                           1 - 0-distances trump other values.

        Returns:
            numpy.ndarray: A confusion matrix of size length(nsam) x length(nsam).

        Raises:
            ValueError: If 'iftrump' is 1 and 'ifresamp' is 2, a ValueError is raised as bootstrapping cannot be used
                when 0-distances trump.
    """

    if expo == 0:
        expo = 'median'
    if not isinstance(expo, str) and expo < 0:
        iftrump = 1
    if iftrump == 1 and ifresamp == 2:
        raise ValueError('Cannot bootstrap resample when 0-distances trump.')

    ncla = len(nsam)  # number of Classes/Tastants
    anear = np.zeros((ncla, ncla))
    if ifresamp == 1:  # relabel resampling
        ind = np.random.permutation(dists.shape[1])
        dists = dists[ind, :][:, ind]
    elif ifresamp == 2:  # a bootstrap resampling
        ndx = []
        nsam2 = np.concatenate([[0], np.cumsum(nsam[:-1])])
        for isam in range(ncla):
            ndx.extend(nsam2[isam] + np.ceil(nsam[isam] * np.random.rand(nsam[isam])).astype(int))
        dists = dists[ndx, :][:, ndx]
    samend = np.cumsum(nsam)
    samstart = np.concatenate([[0], samend[:-1]]) + 1

    av = np.zeros((ncla, dists.shape[1]))  # average distance between trials of the same class
    if iftrump == 0:
        for i in range(ncla):
            if isinstance(expo, str):
                av[i, :] = np.median(dists[samstart[i]:samend[i], :], axis=0)
                # this ensures self-distances (diagonals) are not included in the calculation
                for j in range(samstart[i], samend[i]):
                    av[i, j] = np.median(np.concatenate([dists[samstart[i]:j, j], dists[j + 1:samend[i], j]]))
            else:
                av[i, :] = (np.sum(dists[samstart[i]:samend[i], :] ** expo) / nsam[i]) ** (1 / expo)
                # this ensures self-distances (diagonals) are not included in the calculation
                for j in range(samstart[i], samend[i]):
                    # compute an average of values in subset of rows/columns excluding the diagonal
                    # and average distances from spike-train i to j, except if its negative
                    av[i, j] = (np.sum(np.concatenate([dists[samstart[i]:j, j], dists[j + 1:samend[i], j]]) ** expo) / (
                            nsam[i] - 1)) ** (1 / expo)
    else:
        dists[np.diag_indices_from(dists)] = np.nan
        for i in range(ncla):
            nsm = nsam[i] * np.ones(dists.shape[1])
            nsm[samstart[i] - 1:samend[i]] = nsam[i] - 1  # -1 because python is 0-indexed
            temp = dists[samstart[i] - 1:samend[i], :]  # -1 because python is 0-indexed
            nz = temp < np.finfo(float).eps
            temp = np.zeros_like(temp)
            temp[nz] = 1
            av[i, :] = -np.sum(temp, axis=0) / nsm  # -fraction of zeros in each column
            nz = av[i, :] == 0
            nz_indices = np.where(nz)[0]  # Get the indices where nz is True, columns with no zero values
            ident = np.where((nz_indices >= samstart[i] - 1) & (nz_indices < samend[i]))[0]
            temp = dists[samstart[i] - 1:samend[i], :][:, nz]  # non-zero columns are not all ones
            if isinstance(expo, str):
                md = np.median(temp, axis=0)
                temp = np.sort(temp[:, ident])
                md[ident] = np.median(temp[:nsam[i] - 1, :])
            else:
                # weighted mean distance, train to other classes
                md = (np.sum(temp ** expo, axis=0) / nsam[i]) ** (1 / expo)
                temp = np.sort(temp[:, ident], axis=0)  # compared to matlabs boolean indexing syntax
                md[ident] = (np.sum(temp[:nsam[i] - 1, :] ** expo, axis=0) / (nsam[i] - 1)) ** (1 / expo)
            av[i, nz] = md
    a = np.min(av, axis=0)  # minimum distance between trials of different classes (down columns) - 1 x nspikes
    for i in range(dists.shape[1]):
        icla = np.max(np.where(samstart <= i + 1))  # find the class of the spike-train[i]
        indx = np.where(av[:, i] == a[i])  # find the class of the nearest spike-train to spike-train[i]
        anear[icla, indx[0]] += 1 / len(indx[0])
    return anear.astype(int)  # prosper



import numpy as np
from . import histtpbi as htpbi


def tblxtpbi(ctabl, useall=0):
    """
        Calculates the Treves-Panzeri bias correction to be added to the naive transinformation estimate, in bits.

        Args:
            ctabl (numpy.ndarray): 2-dimensional array of counts.
            useall (int, optional): Use all bins or specific bins based on the parameter. Default is 0.

        Returns:
            float: Treves-Panzeri bias correction estimate in bits.
        """
    h = None
    if useall == 1:
        h = htpbi.histtpbi(np.sum(ctabl, axis=0), 1) + htpbi.histtpbi(np.sum(ctabl, axis=1), 1) - htpbi.histtpbi(ctabl,
                                                                                                                 1)
    elif useall == 0:
        pruned = ctabl[np.where(np.sum(ctabl, axis=1) > 0), :][:, np.where(np.sum(ctabl, axis=0) > 0)]
        h = htpbi.histtpbi(np.sum(pruned, axis=0), 1) + htpbi.histtpbi(np.sum(pruned, axis=1), 1) - htpbi.histtpbi(
            pruned, 1)
    elif useall == -1:
        h = htpbi.histtpbi(np.sum(ctabl, axis=0), 0) + htpbi.histtpbi(np.sum(ctabl, axis=1), 0) - htpbi.histtpbi(ctabl,
                                                                                                                 0)
    return h
